from loguru import logger
from rdflib import RDF
from rdflib import XSD
from rdflib import Graph
from rdflib import Literal
from rdflib import URIRef

from dsp_tools.commands.validate_data.mappers import TRIPLE_OBJECT_TYPE_TO_XSD
from dsp_tools.commands.validate_data.mappers import TRIPLE_PROP_TYPE_TO_IRI_MAPPER
from dsp_tools.commands.validate_data.mappers import VALUE_INFO_TO_RDF_MAPPER
from dsp_tools.commands.validate_data.mappers import VALUE_INFO_TRIPLE_OBJECT_TYPE
from dsp_tools.commands.validate_data.models.rdf_like_data import PropertyObject
from dsp_tools.commands.validate_data.models.rdf_like_data import RdfLikeData
from dsp_tools.commands.validate_data.models.rdf_like_data import RdfLikeResource
from dsp_tools.commands.validate_data.models.rdf_like_data import RdfLikeValue
from dsp_tools.commands.validate_data.models.rdf_like_data import TripleObjectType
from dsp_tools.commands.validate_data.models.rdf_like_data import TriplePropertyType
from dsp_tools.utils.rdflib_constants import DATA
from dsp_tools.utils.xml_parsing.models.parsed_resource import KnoraValueType


def make_data_graph(data: RdfLikeData) -> Graph:
    logger.debug("Creating the RDF data graph.")
    g = Graph()
    for r in data.resources:
        g += _make_one_resource(r)
    return g


def _make_one_resource(res: RdfLikeResource) -> Graph:
    res_iri = DATA[res.res_id]
    g = _make_property_objects_graph(res.property_objects, res_iri)
    for v in res.values:
        g += _make_one_value(v, res_iri)
    if res.asset_value:
        g += _make_one_value(res.asset_value, res_iri)
    return g


def _make_one_value(val: RdfLikeValue, res_iri: URIRef) -> Graph:
    prop_type_info = VALUE_INFO_TO_RDF_MAPPER[val.knora_type]

    val_iri = DATA[val.value_uuid]
    g = _make_property_objects_graph(val.value_metadata, val_iri)
    g.add((res_iri, URIRef(val.user_facing_prop), val_iri))
    g.add((val_iri, RDF.type, prop_type_info.knora_type))
    # The interval values are added in the property objects graph
    if val.knora_type == KnoraValueType.INTERVAL_VALUE:
        return g
    if val.knora_type == KnoraValueType.LINK_VALUE:
        link_val = val.user_facing_value if val.user_facing_value else ""
        triple_object: Literal | URIRef = DATA[link_val]
    else:
        triple_object = _make_one_rdflib_object(val.user_facing_value, VALUE_INFO_TRIPLE_OBJECT_TYPE[val.knora_type])
    g.add((val_iri, prop_type_info.knora_prop, triple_object))
    return g


def _make_property_objects_graph(property_objects: list[PropertyObject], subject_iri: URIRef) -> Graph:
    g = Graph()
    for trpl in property_objects:
        object_val = _make_one_rdflib_object(trpl.object_value, trpl.object_type, trpl.property_type)
        g.add((subject_iri, TRIPLE_PROP_TYPE_TO_IRI_MAPPER[trpl.property_type], object_val))
    return g


def _make_one_rdflib_object(
    object_value: str | None, object_type: TripleObjectType, prop_type: TriplePropertyType | None = None
) -> Literal | URIRef:
    if not object_value:
        return Literal("", datatype=XSD.string)
    if object_type == TripleObjectType.IRI:
        if prop_type == TriplePropertyType.KNORA_STANDOFF_LINK:
            return DATA[object_value]
        return URIRef(object_value)
    return Literal(object_value, datatype=TRIPLE_OBJECT_TYPE_TO_XSD[object_type])



from typing import Any
from typing import Protocol

from dsp_tools.clients.authentication_client import AuthenticationClient


class LegalInfoClient(Protocol):
    """
    Protocol class/interface for the legal info endpoint of the admin API.
    """

    server: str
    project_shortcode: str
    authentication_client: AuthenticationClient

    def post_copyright_holders(self, copyright_holders: list[str]) -> None:
        """Send a list of new copyright holders to the API"""

    def get_licenses_of_a_project(self, enabled_only: bool) -> list[dict[str, Any]]:
        """Get a list of enabled licenses for the project."""

    def enable_unknown_license(self) -> None:
        """Enable the license http://rdfh.ch/licenses/unknown"""



from loguru import logger
from rdflib import RDF
from rdflib import SH
from rdflib import XSD
from rdflib import BNode
from rdflib import Graph
from rdflib import Literal
from rdflib import URIRef
from rdflib.collection import Collection

from dsp_tools.commands.validate_data.models.api_responses import OneList
from dsp_tools.commands.validate_data.models.api_responses import SHACLListInfo
from dsp_tools.utils.rdflib_constants import KNORA_API
from dsp_tools.utils.rdflib_constants import PropertyTypeAlias
from dsp_tools.utils.rdflib_constants import SubjectObjectTypeAlias


def construct_property_shapes(onto: Graph, project_lists: list[OneList]) -> Graph:
    """
    Returns the sh:PropertyShape for the properties in the ontology.

    Args:
        onto: ontology graph
        project_lists: all lists from a project

    Returns:
        Graph with the property shapes
    """
    logger.debug("Constructing property shapes for ontology.")
    g = Graph()
    g += _construct_link_value_shape(onto)
    g += _construct_link_value_node_shape(onto)
    g += _construct_property_type_text_value(onto)
    g += _construct_list_shapes(onto, project_lists)
    g += _construct_seqnum_is_part_of_prop_shape(onto)
    g += _construct_value_type_shapes_to_class_shapes(onto)
    g += _construct_link_value_type_shapes_to_class_shapes(onto)
    return g + _add_property_shapes_to_class_shapes(onto)


def _add_property_shapes_to_class_shapes(onto: Graph) -> Graph:
    logger.debug("Add property shape to resources")
    query_s = """
    PREFIX owl: <http://www.w3.org/2002/07/owl#> 
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#>
    PREFIX knora-api:  <http://api.knora.org/ontology/knora-api/v2#>

    CONSTRUCT {

      ?class a sh:NodeShape ;
             sh:property ?propShapesIRI .

    } WHERE {

      ?class a owl:Class ;
          knora-api:isResourceClass true ;
          knora-api:canBeInstantiated true ;
          rdfs:subClassOf ?restriction .
      ?restriction a owl:Restriction ;          
          owl:onProperty ?propRestriction .

      ?propRestriction knora-api:isEditable true .
      FILTER NOT EXISTS { ?propRestriction knora-api:isLinkValueProperty true }

      BIND(IRI(CONCAT(str(?propRestriction), "_PropShape")) AS ?propShapesIRI)
    }
    """
    if results_graph := onto.query(query_s).graph:
        return results_graph
    return Graph()


def _construct_value_type_shapes_to_class_shapes(onto: Graph) -> Graph:
    logger.debug("Add value type shape to resources")
    query_s = """
    PREFIX owl: <http://www.w3.org/2002/07/owl#> 
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#>
    PREFIX knora-api:  <http://api.knora.org/ontology/knora-api/v2#>

    CONSTRUCT {

      ?shapesIRI a sh:PropertyShape ;
                sh:path ?propRestriction ;
                sh:class ?objectType ;
                sh:message ?objectTypeMessage .

    } WHERE {

      ?class a owl:Class ;
          knora-api:isResourceClass true ;
          knora-api:canBeInstantiated true ;
          rdfs:subClassOf ?restriction .
      ?restriction a owl:Restriction ;          
          owl:onProperty ?propRestriction .

      ?propRestriction knora-api:isEditable true ;
                       knora-api:objectType ?objectType .

      FILTER NOT EXISTS { ?propRestriction knora-api:isLinkProperty true }
      FILTER NOT EXISTS { ?propRestriction knora-api:isLinkValueProperty true }

      BIND(IRI(CONCAT(str(?propRestriction), "_PropShape")) AS ?shapesIRI)
      BIND(CONCAT("This property requires a ", STRAFTER(STR(?objectType), "#")) AS ?objectTypeMessage)
    }
    """
    if results_graph := onto.query(query_s).graph:
        return results_graph
    return Graph()


def _construct_link_value_type_shapes_to_class_shapes(onto: Graph) -> Graph:
    logger.debug("Add value type shape to resources")
    query_s = """
    PREFIX owl: <http://www.w3.org/2002/07/owl#> 
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#>
    PREFIX knora-api:  <http://api.knora.org/ontology/knora-api/v2#>

    CONSTRUCT {

      ?shapesIRI a sh:PropertyShape ;
                 sh:path ?propRestriction ;
                 sh:class knora-api:LinkValue ;
                 sh:message "This property requires a LinkValue" .

    } WHERE {

      ?class a owl:Class ;
          knora-api:isResourceClass true ;
          knora-api:canBeInstantiated true ;
          rdfs:subClassOf ?restriction .
      ?restriction a owl:Restriction ;          
          owl:onProperty ?propRestriction .

      ?propRestriction knora-api:isEditable true ;
                       knora-api:isLinkProperty true .

      BIND(IRI(CONCAT(str(?propRestriction), "_PropShape")) AS ?shapesIRI)
    }
    """
    if results_graph := onto.query(query_s).graph:
        return results_graph
    return Graph()


def _construct_link_value_shape(onto: Graph) -> Graph:
    logger.debug("Constructing LinkValue PropertyShape")
    query_s = """
    PREFIX owl: <http://www.w3.org/2002/07/owl#> 
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    PREFIX api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#>
    PREFIX knora-api:  <http://api.knora.org/ontology/knora-api/v2#>
    PREFIX salsah-gui: <http://api.knora.org/ontology/salsah-gui/v2#>

    CONSTRUCT {

        ?shapesIRI a sh:PropertyShape ;
                   sh:path ?prop ;
                   sh:node ?nodeShapeIRI .

    } WHERE {

        ?prop a owl:ObjectProperty ;
                salsah-gui:guiElement salsah-gui:Searchbox .

        BIND(IRI(CONCAT(str(?prop), "_PropShape")) AS ?shapesIRI)
        BIND(IRI(CONCAT(str(?prop), "_NodeShape")) AS ?nodeShapeIRI)
    }
    """
    if results_graph := onto.query(query_s).graph:
        return results_graph
    return Graph()


def _construct_link_value_node_shape(onto: Graph) -> Graph:
    logger.debug("Constructing LinkValue NodeShape")
    query_s = """
    PREFIX owl: <http://www.w3.org/2002/07/owl#> 
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    PREFIX api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#>
    PREFIX knora-api:  <http://api.knora.org/ontology/knora-api/v2#>
    PREFIX salsah-gui: <http://api.knora.org/ontology/salsah-gui/v2#>

    CONSTRUCT {

        ?nodeShapeIRI a sh:NodeShape ;
            sh:property    [
                                a            sh:PropertyShape ;
                                sh:path      api-shapes:linkValueHasTargetID ;
                                sh:class     ?rangeClass ;
                                sh:message   ?msg ;
                            ] ;
            sh:severity    sh:Violation .

    } WHERE {

        ?prop a owl:ObjectProperty ;
                knora-api:isLinkProperty true ;
                knora-api:objectType ?rangeClass .

        BIND(IRI(CONCAT(str(?prop), "_NodeShape")) AS ?nodeShapeIRI)
        BIND(CONCAT("Range is ", str(?rangeClass), " or a subclass") AS ?msg)
    }
    """
    if results_graph := onto.query(query_s).graph:
        return results_graph
    return Graph()


def _construct_property_type_text_value(onto: Graph) -> Graph:
    property_type_mapper = {
        "salsah-gui:SimpleText": "api-shapes:SimpleTextValue_ClassShape",
        "salsah-gui:Textarea": "api-shapes:TextareaTextValue_ClassShape",
        "salsah-gui:Richtext": "api-shapes:FormattedTextValue_ClassShape",
    }
    g = Graph()
    for object_type, shacl_shape in property_type_mapper.items():
        g += _construct_one_property_type_text_value(onto, object_type, shacl_shape)
    return g


def _construct_one_property_type_text_value(onto: Graph, gui_element: str, shacl_shape: str) -> Graph:
    logger.debug(f"Constructing shape for {gui_element}")
    query_s = """
    PREFIX owl: <http://www.w3.org/2002/07/owl#> 
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    PREFIX api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#>
    PREFIX knora-api:  <http://api.knora.org/ontology/knora-api/v2#>
    PREFIX salsah-gui: <http://api.knora.org/ontology/salsah-gui/v2#>

    CONSTRUCT {

        ?shapesIRI a sh:PropertyShape ;
                   sh:path ?prop ;
                   sh:node %(shacl_shape)s .

    } WHERE {

        ?prop a owl:ObjectProperty ;
                knora-api:objectType knora-api:TextValue ;
                salsah-gui:guiElement %(gui_element)s .

        BIND(IRI(CONCAT(str(?prop), "_PropShape")) AS ?shapesIRI)
    }
    """ % {"gui_element": gui_element, "shacl_shape": shacl_shape}  # noqa: UP031 (printf-string-formatting)
    if results_graph := onto.query(query_s).graph:
        return results_graph
    return Graph()


def _construct_list_shapes(onto: Graph, project_lists: list[OneList]) -> Graph:
    lists_graph = Graph()
    for one_list in project_lists:
        lists_graph += _construct_one_list_node_shape(one_list)
    for one_list in project_lists:
        lists_graph += _construct_one_list_property_shape(onto, one_list)
    return lists_graph


def _construct_one_list_node_shape(one_list: OneList) -> Graph:
    g = Graph()
    list_iri = URIRef(one_list.list_iri)
    g.add((list_iri, RDF.type, SH.NodeShape))
    g.add((list_iri, SH.severity, SH.Violation))
    list_iris = [nd.iri for nd in one_list.nodes]
    node_prop_info = SHACLListInfo(
        list_iri=list_iri,
        sh_path=KNORA_API.listValueAsListNode,
        sh_message=(
            f"A valid node from the list '{one_list.list_name}' must be used with this property "
            f"(input displayed in format 'listName / NodeName')."
        ),
        sh_in=list_iris,
    )
    g += _construct_one_list_property_shape_with_collection(node_prop_info)
    return g


def _construct_one_list_property_shape_with_collection(shacl_info: SHACLListInfo) -> Graph:
    g = Graph()
    collection_bn = BNode()
    collection_literals: list[SubjectObjectTypeAlias] = [Literal(lit, datatype=XSD.string) for lit in shacl_info.sh_in]
    Collection(g, collection_bn, collection_literals)
    prop_shape: list[tuple[PropertyTypeAlias, SubjectObjectTypeAlias]] = [
        (RDF.type, SH.PropertyShape),
        (SH.path, shacl_info.sh_path),
        (URIRef("http://www.w3.org/ns/shacl#in"), collection_bn),
        (SH.message, Literal(shacl_info.sh_message, datatype=XSD.string)),
    ]
    prop_bn: SubjectObjectTypeAlias = BNode()
    for prop, obj in prop_shape:
        g.add((prop_bn, prop, obj))
    g.add((shacl_info.list_iri, SH.property, prop_bn))
    return g


def _construct_one_list_property_shape(onto: Graph, one_list: OneList) -> Graph:
    logger.debug(f"Constructing Collection Shape for ListValue: {one_list.list_iri}")
    query_s = """
    PREFIX owl: <http://www.w3.org/2002/07/owl#> 
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    PREFIX api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#>
    PREFIX knora-api:  <http://api.knora.org/ontology/knora-api/v2#>
    PREFIX salsah-gui: <http://api.knora.org/ontology/salsah-gui/v2#>

    CONSTRUCT {
        ?shapesIRI a sh:PropertyShape ;
                   sh:path ?prop ;
                   sh:node <%(list)s> ;
                   sh:severity sh:Violation .
    } WHERE {
        ?prop a owl:ObjectProperty ;
              knora-api:objectType knora-api:ListValue ;
              salsah-gui:guiAttribute %(guiAttribute)s .

        BIND(IRI(CONCAT(str(?prop), "_PropShape")) AS ?shapesIRI)
    }
    """ % {"guiAttribute": one_list.hlist(), "list": one_list.list_iri}  # noqa: UP031 (printf-string-formatting)
    if results_graph := onto.query(query_s).graph:
        return results_graph
    return Graph()


def _construct_seqnum_is_part_of_prop_shape(onto: Graph) -> Graph:
    logger.debug("Constructing Collection Shape seqnum / isPartOf")
    query_s = """
    PREFIX owl: <http://www.w3.org/2002/07/owl#> 
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    PREFIX dash: <http://datashapes.org/dash#>
    PREFIX api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#>
    PREFIX knora-api:  <http://api.knora.org/ontology/knora-api/v2#>
    PREFIX salsah-gui: <http://api.knora.org/ontology/salsah-gui/v2#>

    CONSTRUCT {
        ?resourceClass sh:property api-shapes:seqnum_PropShape .
    } WHERE {
        ?resourceClass rdfs:subClassOf ?restriction .
        ?restriction a owl:Restriction .
        {
            ?restriction owl:onProperty knora-api:seqnum .
        }
        UNION
        {
            ?restriction owl:onProperty knora-api:isPartOf .
        }
    }
    """
    # The API allows the ontology to declare cardinalities for seqnum without isPartOf and vice versa.
    # Therefore, we have a union query.
    if results_graph := onto.query(query_s).graph:
        return results_graph
    return Graph()



from __future__ import annotations

from dataclasses import dataclass

list_separator = "\n    - "
medium_separator = "\n----------------------------\n"

HTTP_OK = 200


@dataclass(frozen=True)
class AllIIIFUriProblems:
    """Information about all problems with IIIF URIs."""

    problems: list[IIIFUriProblem]

    def get_msg(self) -> str:
        """Get a message describing all problems with the IIIF URIs."""
        msg = "There were problems with the following IIIF URI(s):"
        all_msg = [problem.get_msg() for problem in self.problems]
        return msg + medium_separator + medium_separator.join(all_msg)


@dataclass(frozen=True)
class IIIFUriProblem:
    """Information about a problem with an IIIF URI."""

    uri: str
    regex_has_passed: bool
    status_code: int | None = None
    raised_exception_name: str | None = None
    original_text: str | None = None

    def get_msg(self) -> str:
        """Get a message describing the problem with the IIIF URI."""
        msg = [f"URI: {self.uri}"]
        if self.regex_has_passed:
            msg.append("Passed the internal regex check.")
        else:
            msg.append("Did not pass the internal regex check.")
        if self.status_code == HTTP_OK:
            msg.extend(self._good_status_code_bad_regex_msg())
        elif self.status_code is not None:
            msg.extend(self._bad_status_code_msg())
        else:
            msg.extend(self._exception_msg())
        return list_separator.join(msg)

    def _exception_msg(self) -> list[str]:
        return [
            f"An error occurred during the network call: {self.raised_exception_name}",
            f"Error message: {self.original_text}",
        ]

    def _bad_status_code_msg(self) -> list[str]:
        return [
            "The server did not respond as expected.",
            f"Status code: {self.status_code}",
            f"Response text: {self.original_text}",
        ]

    def _good_status_code_bad_regex_msg(self) -> list[str]:
        return [
            "The URI is correct and the server responded as expected.",
            "Please contact the dsp-tools development team with this information.",
        ]



import os
from collections import defaultdict
from copy import deepcopy
from typing import cast

from dotenv import load_dotenv
from lxml import etree

from dsp_tools.xmllib.internal.circumvent_circular_imports import parse_richtext_as_xml
from dsp_tools.xmllib.internal.constants import DASCH_SCHEMA
from dsp_tools.xmllib.internal.constants import XML_NAMESPACE_MAP
from dsp_tools.xmllib.models.config_options import Permissions
from dsp_tools.xmllib.models.internal.values import BooleanValue
from dsp_tools.xmllib.models.internal.values import ColorValue
from dsp_tools.xmllib.models.internal.values import DateValue
from dsp_tools.xmllib.models.internal.values import DecimalValue
from dsp_tools.xmllib.models.internal.values import GeonameValue
from dsp_tools.xmllib.models.internal.values import IntValue
from dsp_tools.xmllib.models.internal.values import LinkValue
from dsp_tools.xmllib.models.internal.values import ListValue
from dsp_tools.xmllib.models.internal.values import Richtext
from dsp_tools.xmllib.models.internal.values import SimpleText
from dsp_tools.xmllib.models.internal.values import TimeValue
from dsp_tools.xmllib.models.internal.values import UriValue
from dsp_tools.xmllib.models.internal.values import Value

load_dotenv()

PROP_TYPE_LOOKUP = {
    BooleanValue: "boolean",
    ColorValue: "color",
    DateValue: "date",
    DecimalValue: "decimal",
    GeonameValue: "geoname",
    IntValue: "integer",
    LinkValue: "resptr",
    ListValue: "list",
    TimeValue: "time",
    UriValue: "uri",
    Richtext: "richtext",
    SimpleText: "simpletext",
}


def serialise_values(all_values: list[Value]) -> list[etree._Element]:
    """
    Serialise a list of values for a resource.

    Args:
        all_values: List of Values

    Returns:
        list of serialised values
    """
    prop_tuples, type_lookup = _sort_and_group_values(all_values)
    serialised = []
    for prop_name, prop_values in prop_tuples:
        prop_type = type_lookup[prop_name]
        match prop_type:
            case "list":
                serialised.append(_serialise_complete_list_prop(cast(list[ListValue], prop_values), prop_name))
            case "simpletext":
                serialised.append(_serialise_complete_simple_text_prop(cast(list[SimpleText], prop_values), prop_name))
            case "richtext":
                serialised.append(_serialise_complete_richtext_prop(cast(list[Richtext], prop_values), prop_name))
            case _:
                serialised.append(_serialise_complete_generic_prop(prop_values, prop_name, prop_type))
    return serialised


def _sort_and_group_values(all_values: list[Value]) -> tuple[list[tuple[str, list[Value]]], dict[str, str]]:
    prop_groups, type_lookup = _group_properties(all_values)
    prop_tuples = [(prop_name, values) for prop_name, values in prop_groups.items()]
    env_var = str(os.getenv("XMLLIB_SORT_PROPERTIES")).lower()
    if env_var == "true":
        prop_tuples = [
            (prop_name, sorted(prop_values, key=lambda x: x.value)) for (prop_name, prop_values) in prop_tuples
        ]
        prop_tuples = sorted(prop_tuples, key=lambda x: x[0])
    return prop_tuples, type_lookup


def _group_properties(values: list[Value]) -> tuple[dict[str, list[Value]], dict[str, str]]:
    grouped = defaultdict(list)
    name_type_lookup = {}
    for val in values:
        grouped[val.prop_name].append(val)
        name_type_lookup[val.prop_name] = PROP_TYPE_LOOKUP[type(val)]
    return grouped, name_type_lookup


def _serialise_complete_generic_prop(values: list[Value], prop_name: str, prop_type: str) -> etree._Element:
    prop = _serialise_generic_prop(prop_name, prop_type)
    for val in values:
        val_ele = _serialise_generic_element(val, prop_type)
        prop.append(val_ele)
    return prop


def _serialise_generic_prop(prop_name: str, prop_type: str) -> etree._Element:
    return etree.Element(f"{DASCH_SCHEMA}{prop_type}-prop", name=prop_name, nsmap=XML_NAMESPACE_MAP)


def _serialise_generic_element(value: Value, prop_type: str) -> etree._Element:
    attribs = {}
    if value.permissions != Permissions.PROJECT_SPECIFIC_PERMISSIONS:
        attribs["permissions"] = value.permissions.value
    if value.comment is not None:
        attribs["comment"] = str(value.comment)
    ele = etree.Element(f"{DASCH_SCHEMA}{prop_type}", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
    ele.text = str(value.value)
    return ele


def _serialise_complete_list_prop(values: list[ListValue], prop_name: str) -> etree._Element:
    list_name = next(iter(values)).list_name
    prop = etree.Element(f"{DASCH_SCHEMA}list-prop", name=prop_name, list=list_name, nsmap=XML_NAMESPACE_MAP)
    for val in values:
        prop.append(_serialise_generic_element(val, "list"))
    return prop


def _serialise_complete_simple_text_prop(values: list[SimpleText], prop_name: str) -> etree._Element:
    prop = _serialise_generic_prop(prop_name, "text")
    for val in values:
        val_ele = _serialise_generic_element(val, "text")
        val_ele.attrib["encoding"] = "utf8"
        prop.append(val_ele)
    return prop


def _serialise_complete_richtext_prop(values: list[Richtext], prop_name: str) -> etree._Element:
    prop = _serialise_generic_prop(prop_name, "text")
    for val in values:
        val_ele = _serialise_generic_element(val, "text")  # produces wrong text content
        val_ele.attrib["encoding"] = "xml"
        prop.append(_create_richtext_elements_from_string(val, val_ele))  # overwrite the wrong text content
    return prop


def _create_richtext_elements_from_string(value: Richtext, text_element: etree._Element) -> etree._Element:
    new_element = deepcopy(text_element)
    parsed_xml = parse_richtext_as_xml(value.value)
    if not isinstance(parsed_xml, etree._Element):
        raise ValueError(
            "Richtexts should be validated for correct XML syntax when they are created. Apparently this didn't happen."
        )
    new_element.text = parsed_xml.text  # everything before the first child tag
    new_element.extend(list(parsed_xml))  # all (nested) children of the pseudo-xml
    return new_element



from pathlib import Path

import pytest

from dsp_tools.commands.xmlupload.models.lookup_models import XmlReferenceLookups
from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.commands.xmlupload.prepare_xml_input.get_processed_resources import get_processed_resources
from dsp_tools.commands.xmlupload.prepare_xml_input.prepare_xml_input import get_stash_and_upload_order
from dsp_tools.utils.ansi_colors import RESET_TO_DEFAULT
from dsp_tools.utils.ansi_colors import YELLOW
from dsp_tools.utils.xml_parsing.get_parsed_resources import get_parsed_resources
from dsp_tools.utils.xml_parsing.parse_clean_validate_xml import parse_and_clean_xml_file


def test_get_length_ok_resources() -> None:
    test_root = parse_and_clean_xml_file(Path("testdata/xml-data/test-circular-references.xml"))
    parsed_resources = get_parsed_resources(test_root, "https://namespace.ch/")
    permissions_lookup = {"open": Permissions()}
    xml_lookups = XmlReferenceLookups(permissions_lookup, {}, authorships={})
    processed_resources = get_processed_resources(parsed_resources, xml_lookups, is_on_prod_like_server=False)
    _, stash = get_stash_and_upload_order(processed_resources)
    len_standoff = len(stash.standoff_stash.res_2_stash_items)  # type: ignore[union-attr]
    len_resptr = len(stash.link_value_stash.res_2_stash_items)  # type: ignore[union-attr]
    stashed_links = len_standoff + len_resptr
    previous_stash_size = 14
    print_str = (
        f"\n\n---------------------\n"
        f"Total Resources: 63\n"
        f"Previous Stash Size: {previous_stash_size}\n"
        f"Current Stash Size: {stashed_links}"
        f"\n---------------------\n"
    )
    print(YELLOW + print_str + RESET_TO_DEFAULT)
    assert stashed_links <= previous_stash_size


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

from typing import Any

import pandas as pd
import regex

from dsp_tools.error.xmllib_warnings import MessageInfo
from dsp_tools.error.xmllib_warnings_util import emit_xmllib_input_info
from dsp_tools.error.xmllib_warnings_util import emit_xmllib_input_warning


def is_nonempty_value_internal(value: Any) -> bool:
    """
    This is a duplicate of value_checkers.is_nonempty_value(), to avoid circular imports.

    Check if a value is not None-like
    or that its string representation contains at least one of the following characters:

    - ``\\p{S}`` = symbols and special characters
    - ``\\p{P}`` = punctuation
    - ``\\w`` = all Unicode letters, numbers, and _

    Args:
        value: value of any type

    Returns:
        True if the value is not None-like and contains at least one of the above-mentioned characters
    """
    if pd.isna(value):
        return False
    if regex.search(r"[\p{S}\p{P}\w]", str(value), flags=regex.UNICODE):
        return True
    return False


def check_and_warn_potentially_empty_string(
    *, value: Any, res_id: str | None, expected: str, prop_name: str | None = None, field: str | None = None
) -> None:
    """
    If a user str() casts an input before using it in the xmllib we may get `None` values that are not recognised
    as potentially empty.
    This is to be used if the potentially erroneous input is not detectable through any other check.
    For example a color or date value has its specific regex and it is not possible for such a value to slip through.

    Args:
        value: user input
        res_id: Resource ID
        expected: the type of value that is expected
        prop_name: property name if used to check a property
        field: if used to check a non-property field, for example a comment on a value

    Warnings:
        XmllibInputWarning: if it is an empty value or a string only with whitespaces
        XmllibInputInfo: if it is a string containing a string value
            that may be the result of str() casting an empty value
    """
    if not is_nonempty_value_internal(value):
        msg_info = MessageInfo(
            message=f"Your input '{value}' is empty. Please enter a valid {expected}.",
            resource_id=res_id,
            prop_name=prop_name,
            field=field,
        )
        emit_xmllib_input_warning(msg_info)
    else:
        check_and_warn_if_a_string_contains_a_potentially_empty_value(
            value=value, res_id=res_id, prop_name=prop_name, field=field
        )


def check_and_warn_if_a_string_contains_a_potentially_empty_value(
    *, value: Any, res_id: str | None, prop_name: str | None = None, field: str | None = None
) -> None:
    """
    If a user str() casts an input before using it in the xmllib we may get `None` values that are not recognised
    as potentially empty.
    This is to be used if the potentially erroneous input is not detectable through any other check.

    Args:
        value: user input
        res_id: Resource ID
        prop_name: property name if used to check a property
        field: if used to check a non-property field, for example a comment on a value

    Warnings:
        XmllibInputInfo: if it is a string containing a string value
            that may be the result of str() casting an empty value
    """
    empty_val_string = r"^(<NA>|nan|None)$"
    if bool(regex.search(empty_val_string, str(value))):
        type_lookup = {"<NA>": "pd.NA", "nan": "np.nan", "None": "None"}
        msg = (
            f"Your input '{value}' is a string but may be the result of `str({type_lookup[str(value)]})`. "
            f"Please verify that the input is as expected."
        )
        msg_info = MessageInfo(
            message=msg,
            resource_id=res_id,
            prop_name=prop_name,
            field=field,
        )
        emit_xmllib_input_info(msg_info)


def check_and_inform_about_angular_brackets(value: Any, res_id: str | None, prop_name: str | None = None) -> None:
    """
    Checks if a string value contains angular brackets.

    Args:
        value: String value
        res_id: resource id
        prop_name: property name
    """
    if bool(regex.search(r'<([a-zA-Z/"]+|[^\s0-9].*[^\s0-9])>', str(value))):
        msg_info = MessageInfo(
            message=(
                f"Your input '{value}' contains angular brackets. "
                f"Since this is a simpletext, please note that these will not be recognised as formatting "
                f"in the text field, but will be displayed as-is."
            ),
            resource_id=res_id,
            prop_name=prop_name,
        )
        emit_xmllib_input_info(msg_info)



from rdflib import RDF
from rdflib import XSD
from rdflib import BNode
from rdflib import Graph
from rdflib import Literal
from rdflib import URIRef

from dsp_tools.commands.xmlupload.make_rdf_graph.constants import IIIF_URI_VALUE
from dsp_tools.commands.xmlupload.models.bitstream_info import BitstreamInfo
from dsp_tools.commands.xmlupload.models.rdf_models import AbstractFileValue
from dsp_tools.commands.xmlupload.models.rdf_models import FileValueMetadata
from dsp_tools.commands.xmlupload.models.rdf_models import RDFPropTypeInfo
from dsp_tools.utils.rdflib_constants import KNORA_API


def make_iiif_uri_value_graph(iiif_uri: AbstractFileValue, res_node: BNode | URIRef) -> Graph:
    """
    Creates a graph with the IIIF-URI Link

    Args:
        iiif_uri: Information about the IIIF URI
        res_node: Node of the resource

    Returns:
        Graph with the IIIF-URI Value
    """
    return _make_abstract_file_value_graph(
        iiif_uri, IIIF_URI_VALUE, res_node, KNORA_API.stillImageFileValueHasExternalUrl
    )


def make_file_value_graph(
    bitstream_info: BitstreamInfo,
    rdf_prop_type_info: RDFPropTypeInfo,
    file_value_metadata: FileValueMetadata,
    res_node: BNode | URIRef,
) -> Graph:
    """
    Creates a graph with the File Value information.

    Args:
        bitstream_info: Information about the previously uploaded file
        rdf_prop_type_info: the type of value based on the extension
        file_value_metadata: Metadata of the file value
        res_node: Node of the resource

    Returns:
        Graph with the File Value
    """
    internal_filename = bitstream_info.internal_file_name
    file_value = AbstractFileValue(internal_filename, file_value_metadata)
    return _make_abstract_file_value_graph(file_value, rdf_prop_type_info, res_node)


def _make_abstract_file_value_graph(
    file_value: AbstractFileValue,
    type_info: RDFPropTypeInfo,
    res_node: BNode | URIRef,
    has_file_name_prop: URIRef = KNORA_API.fileValueHasFilename,
) -> Graph:
    file_bn = BNode()
    g = _add_metadata(file_bn, file_value.metadata)
    g.add((res_node, type_info.knora_prop, file_bn))
    g.add((file_bn, RDF.type, type_info.knora_type))
    g.add((file_bn, has_file_name_prop, Literal(file_value.value, datatype=XSD.string)))
    return g


def _add_metadata(file_bn: BNode, metadata: FileValueMetadata) -> Graph:
    g = Graph()
    g.add((file_bn, KNORA_API.hasLicense, URIRef(metadata.license_iri)))
    g.add((file_bn, KNORA_API.hasCopyrightHolder, Literal(metadata.copyright_holder, datatype=XSD.string)))
    for auth in metadata.authorships:
        g.add((file_bn, KNORA_API.hasAuthorship, Literal(auth, datatype=XSD.string)))
    if metadata.permissions:
        g.add((file_bn, KNORA_API.hasPermissions, Literal(metadata.permissions, datatype=XSD.string)))
    return g



# mypy: disable-error-code="no-untyped-def"


import pytest
from rdflib import RDF
from rdflib import RDFS
from rdflib import Graph
from rdflib import Literal
from rdflib import URIRef

from dsp_tools.utils.rdflib_constants import KNORA_API
from dsp_tools.utils.rdflib_constants import KNORA_API_STR
from test.e2e.commands.xmlupload.utils import util_get_res_iri_from_label
from test.e2e.commands.xmlupload.utils import util_request_resources_by_class

# ruff: noqa: ARG001 Unused function argument

OPEN_PERMISSIONS = Literal(
    "CR knora-admin:ProjectAdmin|D knora-admin:ProjectMember|V knora-admin:KnownUser,knora-admin:UnknownUser"
)
DOAP_PERMISSIONS = Literal("CR knora-admin:ProjectAdmin|D knora-admin:ProjectMember")

NUMBER_OF_RESOURCE_TRIPLES_WITHOUT_VALUES = 9

"""
Note on these tests:
- The xmlupload code that creates metadata triples (e.g. the permissions) is shared by all resource types.
- FileValue metadata is also created by code that is shared by all file value types, 
  therefore it is not necessary to assert the correct creation for every FileValue type separately.
- The content of values is tested.
- For built-in DSP resources the presence of the expected values is explicitly tested here, 
  but the content of the values is not as that is included in the tests for the values 
  in the designated file of this directory.
"""


@pytest.fixture(scope="module")
def cls_with_everything_graph(
    _xmlupload_minimal_correct, class_with_everything_iri, auth_header, project_iri, creds
) -> Graph:
    return util_request_resources_by_class(class_with_everything_iri, auth_header, project_iri, creds)


class TestResources:
    def test_class_with_everything_all_created(self, cls_with_everything_graph, class_with_everything_iri):
        cls_iri = URIRef(class_with_everything_iri)
        resource_iris = list(cls_with_everything_graph.subjects(RDF.type, cls_iri))
        expected_number_of_resources = 17
        assert len(resource_iris) == expected_number_of_resources

    def test_resource_no_values_assert_triples_present(
        self, cls_with_everything_graph, class_with_everything_iri, project_iri
    ):
        res_iri = util_get_res_iri_from_label(cls_with_everything_graph, "resource_no_values")
        number_of_triples = list(cls_with_everything_graph.triples((res_iri, None, None)))
        assert next(cls_with_everything_graph.objects(res_iri, KNORA_API.attachedToProject)) == URIRef(project_iri)
        assert next(cls_with_everything_graph.objects(res_iri, RDF.type)) == URIRef(class_with_everything_iri)
        assert len(list(cls_with_everything_graph.objects(res_iri, KNORA_API.hasPermissions))) == 1
        assert len(list(cls_with_everything_graph.objects(res_iri, KNORA_API.arkUrl))) == 1
        assert len(list(cls_with_everything_graph.objects(res_iri, KNORA_API.versionArkUrl))) == 1
        assert len(list(cls_with_everything_graph.objects(res_iri, KNORA_API.userHasPermission))) == 1
        assert len(list(cls_with_everything_graph.objects(res_iri, KNORA_API.creationDate))) == 1
        assert len(list(cls_with_everything_graph.objects(res_iri, KNORA_API.attachedToUser))) == 1
        assert len(number_of_triples) == NUMBER_OF_RESOURCE_TRIPLES_WITHOUT_VALUES

    def test_resource_no_permissions_specified(self, cls_with_everything_graph):
        res_iri = util_get_res_iri_from_label(cls_with_everything_graph, "resource_no_values")
        actual_permissions = next(cls_with_everything_graph.objects(res_iri, KNORA_API.hasPermissions))
        assert actual_permissions == DOAP_PERMISSIONS

    def test_resource_with_open_permissions(self, cls_with_everything_graph):
        res_iri = util_get_res_iri_from_label(cls_with_everything_graph, "resource_no_values_open_permissions")
        actual_permissions = next(cls_with_everything_graph.objects(res_iri, KNORA_API.hasPermissions))
        assert actual_permissions == OPEN_PERMISSIONS

    @pytest.mark.usefixtures("_xmlupload_minimal_correct")
    def test_second_onto_class(self, second_onto_iri, onto_iri, auth_header, project_iri, creds):
        cls_iri_str = f"{second_onto_iri}SecondOntoClass"
        g = util_request_resources_by_class(cls_iri_str, auth_header, project_iri, creds)
        resource_iris = list(g.subjects(RDF.type, URIRef(cls_iri_str)))
        expected_number_of_resources = 1
        assert len(resource_iris) == expected_number_of_resources
        res_iri = resource_iris.pop(0)
        assert next(g.objects(res_iri, URIRef(f"{second_onto_iri}testBoolean")))
        assert next(g.objects(res_iri, URIRef(f"{onto_iri}testSimpleText")))

    @pytest.mark.usefixtures("_xmlupload_minimal_correct")
    def test_still_image(self, onto_iri, auth_header, project_iri, creds):
        cls_iri_str = f"{onto_iri}TestStillImageRepresentation"
        g = util_request_resources_by_class(cls_iri_str, auth_header, project_iri, creds)
        resource_iris = list(g.subjects(RDF.type, URIRef(cls_iri_str)))
        expected_number_of_resources = 2
        assert len(resource_iris) == expected_number_of_resources

        image = util_get_res_iri_from_label(g, "image")
        image_triples = list(g.triples((image, None, None)))
        assert len(image_triples) == NUMBER_OF_RESOURCE_TRIPLES_WITHOUT_VALUES + 1
        image_val = next(g.objects(image, KNORA_API.hasStillImageFileValue))
        assert next(g.objects(image_val, RDF.type)) == KNORA_API.StillImageFileValue
        assert next(g.objects(image_val, KNORA_API.hasPermissions)) == DOAP_PERMISSIONS
        assert next(g.objects(image_val, KNORA_API.hasAuthorship)) == Literal("Johannes Nussbaum")
        assert next(g.objects(image_val, KNORA_API.hasCopyrightHolder)) == Literal("DaSCH")
        assert next(g.objects(image_val, KNORA_API.hasLicense)) == URIRef("http://rdfh.ch/licenses/cc-by-4.0")

        iiif_uri = util_get_res_iri_from_label(g, "iiif_uri")
        iiif_uri_triples = list(g.triples((iiif_uri, None, None)))
        assert len(iiif_uri_triples) == NUMBER_OF_RESOURCE_TRIPLES_WITHOUT_VALUES + 1
        iiif_uri_val = next(g.objects(iiif_uri, KNORA_API.hasStillImageFileValue))
        assert next(g.objects(iiif_uri_val, RDF.type)) == KNORA_API.StillImageExternalFileValue
        assert next(g.objects(iiif_uri_val, KNORA_API.hasPermissions)) == OPEN_PERMISSIONS
        assert next(g.objects(iiif_uri_val, KNORA_API.hasAuthorship)) == Literal("Cavanagh, Annie")
        assert next(g.objects(iiif_uri_val, KNORA_API.hasCopyrightHolder)) == Literal("Wellcome Collection")
        assert next(g.objects(iiif_uri_val, KNORA_API.hasLicense)) == URIRef("http://rdfh.ch/licenses/cc-by-nc-4.0")

    @pytest.mark.usefixtures("_xmlupload_minimal_correct")
    def test_audio(self, onto_iri, auth_header, project_iri, creds):
        cls_iri_str = f"{onto_iri}TestAudioRepresentation"
        g = util_request_resources_by_class(cls_iri_str, auth_header, project_iri, creds)
        resource_iris = list(g.subjects(RDF.type, URIRef(cls_iri_str)))
        assert next(g.objects(predicate=RDFS.label)) == Literal("audio")
        expected_number_of_resources = 1
        assert len(resource_iris) == expected_number_of_resources

    @pytest.mark.usefixtures("_xmlupload_minimal_correct")
    def test_video(self, onto_iri, auth_header, project_iri, creds):
        cls_iri_str = f"{onto_iri}TestMovingImageRepresentation"
        g = util_request_resources_by_class(cls_iri_str, auth_header, project_iri, creds)
        resource_iris = list(g.subjects(RDF.type, URIRef(cls_iri_str)))
        assert next(g.objects(predicate=RDFS.label)) == Literal("video")
        expected_number_of_resources = 1
        assert len(resource_iris) == expected_number_of_resources


class TestDspResources:
    @pytest.mark.usefixtures("_xmlupload_minimal_correct")
    def test_region(self, auth_header, project_iri, creds):
        cls_iri_str = f"{KNORA_API_STR}Region"
        g = util_request_resources_by_class(cls_iri_str, auth_header, project_iri, creds)
        resource_iris = list(g.subjects(RDF.type, URIRef(cls_iri_str)))
        expected_number_of_resources = 1
        assert len(resource_iris) == expected_number_of_resources
        res_iri = resource_iris.pop(0)
        res_triples = list(g.triples((res_iri, None, None)))
        assert len(list(g.objects(res_iri, KNORA_API.hasColor))) == 1
        assert len(list(g.objects(res_iri, KNORA_API.isRegionOfValue))) == 1
        assert len(list(g.objects(res_iri, KNORA_API.hasGeometry))) == 1
        assert len(list(g.objects(res_iri, KNORA_API.hasComment))) == 1
        number_of_values = 4
        expected_number_of_triples = NUMBER_OF_RESOURCE_TRIPLES_WITHOUT_VALUES + number_of_values
        assert len(res_triples) == expected_number_of_triples

    @pytest.mark.usefixtures("_xmlupload_minimal_correct")
    def test_link_obj(self, auth_header, project_iri, creds):
        cls_iri_str = f"{KNORA_API_STR}LinkObj"
        g = util_request_resources_by_class(cls_iri_str, auth_header, project_iri, creds)
        resource_iris = list(g.subjects(RDF.type, URIRef(cls_iri_str)))
        expected_number_of_resources = 1
        assert len(resource_iris) == expected_number_of_resources
        res_iri = resource_iris.pop(0)
        res_triples = list(g.triples((res_iri, None, None)))
        assert len(list(g.objects(res_iri, KNORA_API.hasComment))) == 1
        assert len(list(g.objects(res_iri, KNORA_API.hasLinkToValue))) == 2
        number_of_values = 3
        expected_number_of_triples = NUMBER_OF_RESOURCE_TRIPLES_WITHOUT_VALUES + number_of_values
        assert len(res_triples) == expected_number_of_triples

    @pytest.mark.usefixtures("_xmlupload_minimal_correct")
    def test_audio_segment(self, auth_header, project_iri, creds):
        cls_iri_str = f"{KNORA_API_STR}AudioSegment"
        g = util_request_resources_by_class(cls_iri_str, auth_header, project_iri, creds)
        resource_iris = list(g.subjects(RDF.type, URIRef(cls_iri_str)))
        expected_number_of_resources = 1
        assert len(resource_iris) == expected_number_of_resources
        res_iri = resource_iris.pop(0)
        res_triples = list(g.triples((res_iri, None, None)))
        assert len(list(g.objects(res_iri, KNORA_API.isAudioSegmentOfValue))) == 1
        assert len(list(g.objects(res_iri, KNORA_API.hasSegmentBounds))) == 1
        assert len(list(g.objects(res_iri, KNORA_API.hasTitle))) == 1
        assert len(list(g.objects(res_iri, KNORA_API.hasComment))) == 1
        assert len(list(g.objects(res_iri, KNORA_API.hasDescription))) == 1
        assert len(list(g.objects(res_iri, KNORA_API.hasKeyword))) == 1
        number_of_values = 6
        expected_number_of_triples = NUMBER_OF_RESOURCE_TRIPLES_WITHOUT_VALUES + number_of_values
        assert len(res_triples) == expected_number_of_triples

    @pytest.mark.usefixtures("_xmlupload_minimal_correct")
    def test_video_segment(self, auth_header, project_iri, creds):
        cls_iri_str = f"{KNORA_API_STR}VideoSegment"
        g = util_request_resources_by_class(cls_iri_str, auth_header, project_iri, creds)
        resource_iris = list(g.subjects(RDF.type, URIRef(cls_iri_str)))
        expected_number_of_resources = 1
        assert len(resource_iris) == expected_number_of_resources
        res_iri = resource_iris.pop(0)
        res_triples = list(g.triples((res_iri, None, None)))
        assert len(list(g.objects(res_iri, KNORA_API.isVideoSegmentOfValue))) == 1
        assert len(list(g.objects(res_iri, KNORA_API.hasSegmentBounds))) == 1
        assert len(list(g.objects(res_iri, KNORA_API.hasTitle))) == 1
        assert len(list(g.objects(res_iri, KNORA_API.hasComment))) == 1
        assert len(list(g.objects(res_iri, KNORA_API.hasDescription))) == 1
        assert len(list(g.objects(res_iri, KNORA_API.hasKeyword))) == 1
        number_of_values = 6
        expected_number_of_triples = NUMBER_OF_RESOURCE_TRIPLES_WITHOUT_VALUES + number_of_values
        assert len(res_triples) == expected_number_of_triples



# mypy: disable-error-code="no-untyped-def"

from pathlib import Path

import pytest

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.cli.args import ValidateDataConfig
from dsp_tools.cli.args import ValidationSeverity
from dsp_tools.clients.authentication_client import AuthenticationClient
from dsp_tools.clients.authentication_client_live import AuthenticationClientLive
from dsp_tools.commands.validate_data.api_clients import ShaclValidator
from dsp_tools.commands.validate_data.get_user_validation_message import sort_user_problems
from dsp_tools.commands.validate_data.models.input_problems import ProblemType
from dsp_tools.commands.validate_data.models.input_problems import SortedProblems
from dsp_tools.commands.validate_data.query_validation_result import reformat_validation_graph
from dsp_tools.commands.validate_data.validate_data import _get_validation_result
from dsp_tools.commands.validate_data.validate_data import _get_validation_status
from dsp_tools.commands.validate_data.validate_data import _prepare_data_for_validation_from_file

# ruff: noqa: ARG001 Unused function argument


CONFIG = ValidateDataConfig(Path(), None, ValidationSeverity.INFO, is_on_prod_server=False)


@pytest.fixture(scope="module")
def authentication(creds: ServerCredentials) -> AuthenticationClient:
    auth = AuthenticationClientLive(server=creds.server, email=creds.user, password=creds.password)
    return auth


@pytest.fixture(scope="module")
def no_violations_with_warnings(
    create_generic_project, authentication, shacl_validator: ShaclValidator
) -> SortedProblems:
    file = Path("testdata/validate-data/generic/no_violations_with_warnings.xml")
    graphs, used_iris = _prepare_data_for_validation_from_file(file, authentication)
    report = _get_validation_result(graphs, shacl_validator, CONFIG)
    reformatted = reformat_validation_graph(report)
    return sort_user_problems(reformatted)


@pytest.fixture(scope="module")
def no_violations_with_info(create_generic_project, authentication, shacl_validator: ShaclValidator) -> SortedProblems:
    file = Path("testdata/validate-data/generic/no_violations_with_info.xml")
    graphs, used_iris = _prepare_data_for_validation_from_file(file, authentication)
    report = _get_validation_result(graphs, shacl_validator, CONFIG)
    reformatted = reformat_validation_graph(report)
    return sort_user_problems(reformatted)


class TestGetCorrectValidationResult:
    def test_no_violations_with_warnings_not_on_prod(self, no_violations_with_warnings):
        result = _get_validation_status(no_violations_with_warnings, is_on_prod=False)
        assert result is True

    def test_no_violations_with_warnings_on_prod(self, no_violations_with_warnings):
        result = _get_validation_status(no_violations_with_warnings, is_on_prod=True)
        assert result is False

    def test_no_violations_with_info_not_on_prod(self, no_violations_with_info):
        result = _get_validation_status(no_violations_with_info, is_on_prod=False)
        assert result is True

    def test_no_violations_with_info_on_prod(self, no_violations_with_info):
        result = _get_validation_status(no_violations_with_info, is_on_prod=True)
        assert result is True


class TestSortedProblems:
    def test_no_violations_with_warnings_problems(self, no_violations_with_warnings):
        expected_warnings = [
            # each type of missing legal info (authorship, copyright, license) produces one violation
            ("archive_no_legal_info", ProblemType.GENERIC),
            ("archive_no_legal_info", ProblemType.GENERIC),
            ("archive_no_legal_info", ProblemType.GENERIC),
            ("iiif_no_legal_info", ProblemType.GENERIC),
            ("iiif_no_legal_info", ProblemType.GENERIC),
            ("iiif_no_legal_info", ProblemType.GENERIC),
            ("image_no_legal_info", ProblemType.GENERIC),
            ("image_no_legal_info", ProblemType.GENERIC),
            ("image_no_legal_info", ProblemType.GENERIC),
        ]
        sorted_warnings = sorted(no_violations_with_warnings.user_warnings, key=lambda x: x.res_id)
        assert not no_violations_with_warnings.unique_violations
        assert len(no_violations_with_warnings.user_warnings) == len(expected_warnings)
        assert not no_violations_with_warnings.user_info
        assert not no_violations_with_warnings.unexpected_shacl_validation_components
        for one_result, expected_info in zip(sorted_warnings, expected_warnings):
            assert one_result.problem_type == expected_info[1]
            assert one_result.res_id == expected_info[0]

    def test_no_violations_with_info(self, no_violations_with_info):
        all_expected_info = [
            ("duplicate_iiif_1", ProblemType.FILE_DUPLICATE),
            ("duplicate_iiif_2", ProblemType.FILE_DUPLICATE),
            ("duplicate_still_image_1", ProblemType.FILE_DUPLICATE),
            ("duplicate_still_image_2", ProblemType.FILE_DUPLICATE),
            ("link_to_resource_in_db", ProblemType.INEXISTENT_LINKED_RESOURCE),
            ("triplicate_archive_1", ProblemType.FILE_DUPLICATE),
            ("triplicate_archive_2", ProblemType.FILE_DUPLICATE),
            ("triplicate_archive_3", ProblemType.FILE_DUPLICATE),
        ]
        sorted_info = sorted(no_violations_with_info.user_info, key=lambda x: x.res_id)
        assert not no_violations_with_info.unique_violations
        assert not no_violations_with_info.user_warnings
        assert len(no_violations_with_info.user_info) == len(all_expected_info)
        assert not no_violations_with_info.unexpected_shacl_validation_components
        for one_result, expected_info in zip(sorted_info, all_expected_info):
            assert one_result.problem_type == expected_info[1]
            assert one_result.res_id == expected_info[0]


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

from enum import StrEnum


class License(StrEnum): ...


class LicenseRecommended:
    """
    Recommended licenses:

    - `DSP`: Licenses created and curated by DaSCH, [see `DSP` for details.](https://docs.dasch.swiss/latest/DSP-TOOLS/xmllib-api-reference/licenses/recommended/#xmllib.models.licenses.recommended.DSP)
    - `CC`: Creative Commons licenses, [see `CC` for details.](https://docs.dasch.swiss/latest/DSP-TOOLS/xmllib-api-reference/licenses/recommended/#xmllib.models.licenses.recommended.CC)

    Tip:
        Use the helper function [find_license_in_string()](https://docs.dasch.swiss/latest/DSP-TOOLS/xmllib-api-reference/helpers/#xmllib.helpers.find_license_in_string)
        to parse a license from a string.

    Examples:
        ```python
        # adding a Creative Commons license to a file
        resource = resource.add_file(
            filename="images/cat.jpg",
            license=xmllib.LicenseRecommended.CC.BY,
            copyright_holder="Meow University",
            authorship=["Kitty Meow"],
        )
        ```

        ```python
        # adding a DSP license to a file
        resource = resource.add_file(
            filename="images/dog.jpg",
            license=xmllib.LicenseRecommended.DSP.PUBLIC_DOMAIN,
            copyright_holder="Bark University",
            authorship=["Bark McDog"],
        )
        ```
    """

    CC: type[CC]
    DSP: type[DSP]


class CC(License):
    """
    Pre-defined and recommended [Creative Commons licenses:](https://creativecommons.org/share-your-work/)

    - `BY`: [Attribution 4.0](https://creativecommons.org/licenses/by/4.0/)
    - `BY_SA`: [Attribution-ShareAlike 4.0](https://creativecommons.org/licenses/by-sa/4.0/)
    - `BY_NC`: [Attribution-NonCommercial 4.0](https://creativecommons.org/licenses/by-nc/4.0/)
    - `BY_NC_SA`: [Attribution-NonCommercial-ShareAlike 4.0](https://creativecommons.org/licenses/by-nc-sa/4.0/)
    - `BY_ND`: [Attribution-NoDerivatives 4.0](https://creativecommons.org/licenses/by-nd/4.0/)
    - `BY_NC_ND`: [Attribution-NonCommercial-NoDerivatives 4.0](https://creativecommons.org/licenses/by-nc-nd/4.0/)

    Tip:
        Use the helper function [find_license_in_string()](https://docs.dasch.swiss/latest/DSP-TOOLS/xmllib-api-reference/helpers/#xmllib.helpers.find_license_in_string)
        to parse a license from a string.

    Examples:
        ```python
        # adding a Creative Commons license to a file
        resource = resource.add_file(
            filename="images/dog.jpg",
            license=xmllib.LicenseRecommended.CC.BY_NC_ND,
            copyright_holder="Bark University",
            authorship=["Bark McDog"],
        )
        ```
    """

    BY = "http://rdfh.ch/licenses/cc-by-4.0"
    BY_SA = "http://rdfh.ch/licenses/cc-by-sa-4.0"
    BY_NC = "http://rdfh.ch/licenses/cc-by-nc-4.0"
    BY_NC_SA = "http://rdfh.ch/licenses/cc-by-nc-sa-4.0"
    BY_ND = "http://rdfh.ch/licenses/cc-by-nd-4.0"
    BY_NC_ND = "http://rdfh.ch/licenses/cc-by-nc-nd-4.0"


class DSP(License):
    """
    Pre-defined and recommended licenses created and curated by DaSCH:

    - `AI_GENERATED`: AI-Generated Content - Not Protected by Copyright
    - `UNKNOWN`: Unknown License - Ask Copyright Holder for Permission
    - `PUBLIC_DOMAIN`: Public Domain - Not Protected by Copyright

    Examples:
        ```python
        # adding a DSP license to a file
        resource = resource.add_file(
            filename="images/cat.jpg",
            license=xmllib.LicenseRecommended.DSP.PUBLIC_DOMAIN,
            copyright_holder="Meow University",
            authorship=["Kitty Meow"],
        )
        ```
    """

    AI_GENERATED = "http://rdfh.ch/licenses/ai-generated"
    UNKNOWN = "http://rdfh.ch/licenses/unknown"
    PUBLIC_DOMAIN = "http://rdfh.ch/licenses/public-domain"


LicenseRecommended.CC = CC
LicenseRecommended.DSP = DSP



from __future__ import annotations

import os
from collections.abc import Collection
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Union
from uuid import uuid4

import pandas as pd
from dotenv import load_dotenv
from loguru import logger
from lxml import etree

from dsp_tools.utils.ansi_colors import BOLD_GREEN
from dsp_tools.utils.ansi_colors import BOLD_RED
from dsp_tools.utils.ansi_colors import RESET_TO_DEFAULT
from dsp_tools.utils.xml_parsing.parse_clean_validate_xml import validate_root_emit_user_message
from dsp_tools.xmllib.internal.constants import DASCH_SCHEMA
from dsp_tools.xmllib.internal.constants import XML_NAMESPACE_MAP
from dsp_tools.xmllib.internal.serialise_resource import serialise_resources
from dsp_tools.xmllib.models.config_options import Permissions
from dsp_tools.xmllib.models.dsp_base_resources import AudioSegmentResource
from dsp_tools.xmllib.models.dsp_base_resources import LinkResource
from dsp_tools.xmllib.models.dsp_base_resources import RegionResource
from dsp_tools.xmllib.models.dsp_base_resources import VideoSegmentResource
from dsp_tools.xmllib.models.internal.file_values import AuthorshipLookup
from dsp_tools.xmllib.models.internal.serialise_permissions import XMLPermissions
from dsp_tools.xmllib.models.res import Resource

type AnyResource = Union[Resource, RegionResource, LinkResource, VideoSegmentResource, AudioSegmentResource]

load_dotenv()

# ruff: noqa: D101


@dataclass
class XMLRoot:
    shortcode: str
    default_ontology: str
    resources: list[AnyResource] = field(default_factory=list)

    @staticmethod
    def create_new(shortcode: str, default_ontology: str) -> XMLRoot:
        """
        Create a new XML root, for one file.

        Args:
            shortcode: project shortcode
            default_ontology: name of the default ontology

        Returns:
            Instance of `XMLRoot`

        Examples:
            ```python
            root = xmllib.XMLRoot.create_new(
                shortcode="0000",
                default_ontology="onto"
            )
            ```
        """
        return XMLRoot(shortcode=shortcode, default_ontology=default_ontology)

    def add_resource(self, resource: AnyResource) -> XMLRoot:
        """
        Add one resource to the root.

        Args:
            resource: any one of:
                    `Resource`,
                    `RegionResource`,
                    `LinkResource`,
                    `VideoSegmentResource`,
                    `AudioSegmentResource`

        Returns:
            The original XMLRoot, with the added resource

        Examples:
            ```python
            resource = xmllib.Resource.create_new(
                res_id="ID", restype=":ResourceType", label="label"
            )

            root = root.add_resource(resource)
            ```
        """
        self.resources.append(resource)
        return self

    def add_resource_multiple(self, resources: Collection[AnyResource]) -> XMLRoot:
        """
        Add a list of resources to the root.

        Args:
            resources: a list of:
                    `Resource`,
                    `RegionResource`,
                    `LinkResource`,
                    `VideoSegmentResource`,
                    `AudioSegmentResource`
                    The types of the resources may be mixed.

        Returns:
            The original XMLRoot, with the added resource

        Examples:
            ```python
            resource_1 = xmllib.Resource.create_new(
                res_id="ID_1", restype=":ResourceType", label="label 1"
            )
            resource_2 = xmllib.Resource.create_new(
                res_id="ID_2", restype=":ResourceType", label="label 2"
            )

            root = root.add_resource_multiple([resource_1, resource_2])
            ```
        """
        self.resources.extend(resources)
        return self

    def add_resource_optional(self, resource: AnyResource | None) -> XMLRoot:
        """
        If the resource is not None, add it to the XMLRoot, otherwise return the XMLRoot unchanged.

        Args:
            resource: any one of:
                    `Resource`,
                    `RegionResource`,
                    `LinkResource`,
                    `VideoSegmentResource`,
                    `AudioSegmentResource`

        Returns:
            The original XMLRoot, with the added value if it was not empty. Else the unchanged original XMLRoot.

        Examples:
            ```python
            resource = xmllib.Resource.create_new(
                res_id="ID", restype=":ResourceType", label="label"
            )

            root = root.add_resource_optional(resource)
            ```

            ```python
            root = root.add_resource_optional(None)
            ```
        """
        if resource:
            self.resources.append(resource)
        return self

    def write_file(self, filepath: str | Path, default_permissions: Permissions | None = None) -> None:
        """
        Write the finished XML to a file.

        Args:
            filepath: where to save the file
            default_permissions: permissions to overwrite `Permissions.PROJECT_SPECIFIC_PERMISSIONS`

        Attention:
            If the parameter `default_permission` is not filled
             and no permissions were assigned to values and resources,
             the default permissions (DOAP) on DSP will be applied.
             If the default permissions (DOAP) are not changed via DSP-API,
             only project members and admins have view permissions.
             In the near future,
             custom DOAPs will be able to be set in the project JSON file (ongoing DSP-TOOLS project).
             At that time this parameter will be deprecated.

             **This will not influence individually assigned permissions, for example
             `Permissions.RESTRICTED` will stay restricted even if your default is set to `Permissions.OPEN`.**

        Warning:
            if the XML is not valid according to the schema

        Examples:
            ```python
            root.write_file("xml_file_name.xml")
            ```

            ```python
            # To overwrite `Permissions.PROJECT_SPECIFIC_PERMISSIONS`
            root.write_file("xml_file_name.xml", Permissions.OPEN)
            ```
        """
        root = self.serialise(default_permissions)

        # The logging is only configured when using the CLI entry point.
        # If this is not disabled, then the statements will also be printed out on the terminal.
        logger.disable("dsp_tools")
        validate_root_emit_user_message(root, Path(filepath).parent)

        etree.indent(root, space="    ")
        xml_string = etree.tostring(
            root,
            encoding="unicode",
            pretty_print=True,
            doctype='<?xml version="1.0" encoding="UTF-8"?>',
        )
        with open(filepath, "w", encoding="utf-8") as f:
            f.write(xml_string)
            print(f"The XML file was successfully saved to {filepath}.")
        if file_path := os.getenv("XMLLIB_WARNINGS_CSV_SAVEPATH"):
            df = pd.read_csv(file_path)
            if len(df) > 0:
                msg = f"{len(df)} warnings occurred, please consult '{file_path}' for details."
                print(BOLD_RED, msg, RESET_TO_DEFAULT)
            else:
                msg = "No warnings occurred during the runtime."
                print(BOLD_GREEN, msg, RESET_TO_DEFAULT)

    def serialise(self, default_permissions: Permissions | None = None) -> etree._Element:
        """
        Create an `lxml.etree._Element` with the information in the root.
        If you wish to create a file, we recommend using the `write_file` method instead.

        Args:
            default_permissions: permissions to overwrite `Permissions.PROJECT_SPECIFIC_PERMISSIONS`

        Returns:
            The `XMLRoot` serialised as XML
        """
        root = self._make_root()
        permissions = XMLPermissions().serialise()
        root.extend(permissions)
        author_lookup = _make_authorship_lookup(self.resources)
        authorship = _serialise_authorship(author_lookup.lookup)
        root.extend(authorship)
        serialised_resources = serialise_resources(self.resources, author_lookup, default_permissions)
        root.extend(serialised_resources)
        return root

    def _make_root(self) -> etree._Element:
        schema_url = (
            "https://raw.githubusercontent.com/dasch-swiss/dsp-tools/main/src/dsp_tools/resources/schema/data.xsd"
        )
        schema_location_key = str(etree.QName("http://www.w3.org/2001/XMLSchema-instance", "schemaLocation"))
        schema_location_value = f"https://dasch.swiss/schema {schema_url}"
        return etree.Element(
            f"{DASCH_SCHEMA}knora",
            attrib={
                schema_location_key: schema_location_value,
                "shortcode": self.shortcode,
                "default-ontology": self.default_ontology,
            },
            nsmap=XML_NAMESPACE_MAP,
        )


def _make_authorship_lookup(resources: list[AnyResource]) -> AuthorshipLookup:
    filtered_resources = [x for x in resources if isinstance(x, Resource)]
    file_vals = [x.file_value for x in filtered_resources if x.file_value]
    authors = {x.metadata.authorship for x in file_vals if x.metadata.authorship}
    sorted_authors = sorted(authors)
    env_var = str(os.getenv("XMLLIB_AUTHORSHIP_ID_WITH_INTEGERS")).lower()
    if env_var == "true":
        auth_nums: list[int | str] = list(range(1, len(sorted_authors) + 1))
    else:
        auth_nums = [str(uuid4()) for _ in range(len(sorted_authors))]
    auth_ids = [f"authorship_{x}" for x in auth_nums]
    lookup = dict(zip(sorted_authors, auth_ids))
    return AuthorshipLookup(lookup)


def _serialise_authorship(authorship_lookup: dict[tuple[str, ...], str]) -> list[etree._Element]:
    to_serialise = [(auth, id_) for auth, id_ in authorship_lookup.items()]
    to_serialise = sorted(to_serialise, key=lambda x: x[0])
    return [_make_one_authorship_element(auth, id_) for auth, id_ in to_serialise]


def _make_one_authorship_element(authors: tuple[str, ...], author_id: str) -> etree._Element:
    def _make_one_author(author: str) -> etree._Element:
        ele = etree.Element(f"{DASCH_SCHEMA}author", nsmap=XML_NAMESPACE_MAP)
        ele.text = author
        return ele

    authorship_ele = etree.Element(f"{DASCH_SCHEMA}authorship", attrib={"id": author_id}, nsmap=XML_NAMESPACE_MAP)
    all_authors = [_make_one_author(x) for x in authors]
    authorship_ele.extend(all_authors)
    return authorship_ele



from __future__ import annotations

from dataclasses import dataclass

from rdflib import URIRef

from dsp_tools.commands.xmlupload.iri_resolver import IriResolver
from dsp_tools.commands.xmlupload.models.permission import Permissions


@dataclass
class XmlReferenceLookups:
    permissions: dict[str, Permissions]
    listnodes: dict[tuple[str, str], str]
    authorships: dict[str, list[str]]


@dataclass
class IRILookups:
    project_iri: URIRef
    id_to_iri: IriResolver



# mypy: disable-error-code="no-untyped-def"

from pathlib import Path
from typing import Never
from typing import assert_never
from typing import cast

import pytest
from rdflib import BNode
from rdflib import URIRef

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.cli.args import ValidateDataConfig
from dsp_tools.cli.args import ValidationSeverity
from dsp_tools.clients.authentication_client import AuthenticationClient
from dsp_tools.clients.authentication_client_live import AuthenticationClientLive
from dsp_tools.commands.validate_data.api_clients import ShaclValidator
from dsp_tools.commands.validate_data.get_user_validation_message import sort_user_problems
from dsp_tools.commands.validate_data.models.input_problems import ProblemType
from dsp_tools.commands.validate_data.models.input_problems import UnknownClassesInData
from dsp_tools.commands.validate_data.models.validation import DetailBaseInfo
from dsp_tools.commands.validate_data.models.validation import RDFGraphs
from dsp_tools.commands.validate_data.models.validation import ValidationReportGraphs
from dsp_tools.commands.validate_data.query_validation_result import _extract_base_info_of_resource_results
from dsp_tools.commands.validate_data.query_validation_result import reformat_validation_graph
from dsp_tools.commands.validate_data.validate_data import _check_for_unknown_resource_classes
from dsp_tools.commands.validate_data.validate_data import _get_validation_result
from dsp_tools.commands.validate_data.validate_data import _get_validation_status
from dsp_tools.commands.validate_data.validate_data import _prepare_data_for_validation_from_file

# ruff: noqa: ARG001 Unused function argument


CONFIG = ValidateDataConfig(Path(), None, ValidationSeverity.INFO, is_on_prod_server=False)


@pytest.fixture(scope="module")
def authentication(creds: ServerCredentials) -> AuthenticationClient:
    auth = AuthenticationClientLive(server=creds.server, email=creds.user, password=creds.password)
    return auth


@pytest.fixture(scope="module")
def unknown_classes_graphs(create_generic_project, authentication) -> tuple[RDFGraphs, set[str]]:
    file = Path("testdata/validate-data/generic/unknown_classes.xml")
    graphs, used_iris = _prepare_data_for_validation_from_file(file, authentication)
    return graphs, used_iris


def test_check_for_unknown_resource_classes(unknown_classes_graphs: tuple[RDFGraphs, set[str]]) -> None:
    graphs, used_iris = unknown_classes_graphs
    result = _check_for_unknown_resource_classes(graphs, used_iris)
    assert isinstance(result, UnknownClassesInData)
    expected = {"onto:NonExisting", "unknown:ClassWithEverything", "unknownClass"}
    assert result.unknown_classes == expected


@pytest.fixture(scope="module")
def unique_value_violation(
    create_generic_project, authentication, shacl_validator: ShaclValidator
) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/unique_value_violation.xml")
    graphs, _ = _prepare_data_for_validation_from_file(file, authentication)
    return _get_validation_result(graphs, shacl_validator, CONFIG)


@pytest.fixture(scope="module")
def file_value_violation(
    create_generic_project, authentication, shacl_validator: ShaclValidator
) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/file_value_violation.xml")
    graphs, _ = _prepare_data_for_validation_from_file(file, authentication)
    return _get_validation_result(graphs, shacl_validator, CONFIG)


@pytest.fixture(scope="module")
def dsp_inbuilt_violation(
    create_generic_project, authentication, shacl_validator: ShaclValidator
) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/dsp_inbuilt_violation.xml")
    graphs, _ = _prepare_data_for_validation_from_file(file, authentication)
    return _get_validation_result(graphs, shacl_validator, CONFIG)


@pytest.fixture(scope="module")
def cardinality_violation(
    create_generic_project, authentication, shacl_validator: ShaclValidator
) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/cardinality_violation.xml")
    graphs, _ = _prepare_data_for_validation_from_file(file, authentication)
    return _get_validation_result(graphs, shacl_validator, CONFIG)


@pytest.fixture(scope="module")
def content_violation(
    create_generic_project, authentication, shacl_validator: ShaclValidator
) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/content_violation.xml")
    graphs, _ = _prepare_data_for_validation_from_file(file, authentication)
    return _get_validation_result(graphs, shacl_validator, CONFIG)


@pytest.fixture(scope="module")
def value_type_violation(
    create_generic_project, authentication, shacl_validator: ShaclValidator
) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/value_type_violation.xml")
    graphs, _ = _prepare_data_for_validation_from_file(file, authentication)
    return _get_validation_result(graphs, shacl_validator, CONFIG)


@pytest.fixture(scope="module")
def every_violation_combination_once(
    create_generic_project, authentication, shacl_validator: ShaclValidator
) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/every_violation_combination_once.xml")
    graphs, _ = _prepare_data_for_validation_from_file(file, authentication)
    return _get_validation_result(graphs, shacl_validator, CONFIG)


def test_cardinality_violation(cardinality_violation: ValidationReportGraphs) -> None:
    assert not cardinality_violation.conforms


def test_reformat_cardinality_violation(cardinality_violation: ValidationReportGraphs) -> None:
    expected_info_tuples = [
        ("id_card_one", ProblemType.MIN_CARD),
        ("id_closed_constraint", ProblemType.NON_EXISTING_CARD),
        ("id_max_card", ProblemType.MAX_CARD),
        ("id_min_card", ProblemType.MIN_CARD),
        ("super_prop_no_card", ProblemType.NON_EXISTING_CARD),
    ]
    result = reformat_validation_graph(cardinality_violation)
    sorted_problems = sort_user_problems(result)
    assert len(sorted_problems.unique_violations) == len(expected_info_tuples)
    assert not sorted_problems.user_warnings
    assert not sorted_problems.user_info
    assert not sorted_problems.unexpected_shacl_validation_components
    alphabetically_sorted = sorted(result.problems, key=lambda x: x.res_id)
    for one_result, expected_info in zip(alphabetically_sorted, expected_info_tuples):
        assert one_result.res_id == expected_info[0]
        assert one_result.problem_type == expected_info[1]
    assert not _get_validation_status(sorted_problems, is_on_prod=True)
    assert not _get_validation_status(sorted_problems, is_on_prod=False)


def test_content_violation(content_violation: ValidationReportGraphs) -> None:
    assert not content_violation.conforms


def test_reformat_content_violation(content_violation: ValidationReportGraphs) -> None:
    expected_info_tuples = [
        (
            "comment_on_value_empty",
            "onto:testUriValue",
            "The comment on the value must be a non-empty string",
        ),
        (
            "comment_on_value_whitespace",
            "onto:testUriValue",
            "The comment on the value must be a non-empty string",
        ),
        ("empty_label", "rdfs:label", "The label must be a non-empty string without newlines."),
        ("empty_text_rich", "onto:testRichtext", "The value must be a non-empty string"),
        ("empty_text_simple", "onto:testTextarea", "The value must be a non-empty string"),
        ("geoname_not_number", "onto:testGeoname", "The value must be a valid geoname code"),
        ("label_with_newline", "rdfs:label", "The label must be a non-empty string without newlines."),
        ("link_target_non_existent", "onto:testHasLinkTo", "other"),
        ("link_target_wrong_class", "onto:testHasLinkToCardOneResource", "id_9_target"),
        (
            "list_name_attrib_empty",
            "onto:testListProp",
            (
                "A valid node from the list 'firstList' must be used with this property "
                "(input displayed in format 'listName / NodeName')."
            ),
        ),
        (
            "list_name_non_existent",
            "onto:testListProp",
            (
                "A valid node from the list 'firstList' must be used with this property "
                "(input displayed in format 'listName / NodeName')."
            ),
        ),
        (
            "list_node_non_existent",
            "onto:testListProp",
            (
                "A valid node from the list 'firstList' must be used with this property "
                "(input displayed in format 'listName / NodeName')."
            ),
        ),
        (
            "richtext_standoff_link_nonexistent",
            "hasStandoffLinkTo",
            "A stand-off link must target an existing resource.",
        ),
        (
            "simple_text_with_newlines",
            "onto:testSimpleText",
            "The value must be a non-empty string without newlines.",
        ),
        ("text_only_whitespace_simple", "onto:testTextarea", "The value must be a non-empty string"),
        (
            "wrong_list_used",
            "onto:testListProp",
            (
                "A valid node from the list 'firstList' must be used with this property "
                "(input displayed in format 'listName / NodeName')."
            ),
        ),
    ]
    result = reformat_validation_graph(content_violation)
    sorted_problems = sort_user_problems(result)
    assert len(sorted_problems.unique_violations) == len(expected_info_tuples)
    assert not sorted_problems.user_warnings
    assert not sorted_problems.user_info
    assert not sorted_problems.unexpected_shacl_validation_components
    assert not result.unexpected_results
    assert len(result.problems) == len(expected_info_tuples)
    alphabetically_sorted = sorted(sorted_problems.unique_violations, key=lambda x: x.res_id)
    for one_result, expected_info in zip(alphabetically_sorted, expected_info_tuples):
        assert one_result.res_id == expected_info[0]
        assert one_result.prop_name == expected_info[1]
        if one_result.problem_type == ProblemType.INPUT_REGEX:
            assert one_result.expected == expected_info[2]
        elif one_result.problem_type == ProblemType.GENERIC:
            assert one_result.message == expected_info[2]
        elif one_result.problem_type == ProblemType.LINK_TARGET_TYPE_MISMATCH:
            assert one_result.input_value == expected_info[2]
        elif one_result.problem_type == ProblemType.INEXISTENT_LINKED_RESOURCE:
            assert one_result.input_value == expected_info[2]
        else:
            nev: Never = cast(Never, one_result.problem_type)
            assert_never(nev)
    assert not _get_validation_status(sorted_problems, is_on_prod=True)
    assert not _get_validation_status(sorted_problems, is_on_prod=False)


def test_value_type_violation(value_type_violation: ValidationReportGraphs) -> None:
    assert not value_type_violation.conforms


def test_reformat_value_type_violation(value_type_violation: ValidationReportGraphs) -> None:
    expected_info_tuples = [
        ("bool_wrong_value_type", "This property requires a BooleanValue", "onto:testBoolean"),
        ("color_wrong_value_type", "This property requires a ColorValue", "onto:testColor"),
        ("date_wrong_value_type", "This property requires a DateValue", "onto:testSubDate1"),
        ("decimal_wrong_value_type", "This property requires a DecimalValue", "onto:testDecimalSimpleText"),
        ("geoname_wrong_value_type", "This property requires a GeonameValue", "onto:testGeoname"),
        ("integer_wrong_value_type", "This property requires a IntValue", "onto:testIntegerSimpleText"),
        ("is_date_should_be_simpletext", "This property requires a TextValue", "onto:testTextarea"),
        ("is_link_should_be_text", "TextValue without formatting", "onto:testTextarea"),
        ("is_text_should_be_integer", "This property requires a IntValue", "onto:testIntegerSpinbox"),
        ("link_wrong_value_type", "This property requires a LinkValue", "onto:testHasLinkTo"),
        ("list_wrong_value_type", "This property requires a ListValue", "onto:testListProp"),
        ("richtext_wrong_value_type", "TextValue with formatting", "onto:testRichtext"),
        ("simpletext_wrong_value_type", "TextValue without formatting", "onto:testTextarea"),
        ("time_wrong_value_type", "This property requires a TimeValue", "onto:testTimeValue"),
        ("uri_wrong_value_type", "This property requires a UriValue", "onto:testUriValue"),
    ]
    result = reformat_validation_graph(value_type_violation)
    # "is_link_should_be_text" gives two types of validation errors, this function removes the duplicates
    sorted_problems = sort_user_problems(result)
    assert len(sorted_problems.unique_violations) == len(expected_info_tuples)
    assert not sorted_problems.user_warnings
    assert not sorted_problems.user_info
    assert not sorted_problems.unexpected_shacl_validation_components
    alphabetically_sorted = sorted(sorted_problems.unique_violations, key=lambda x: x.res_id)
    for one_result, expected_info in zip(alphabetically_sorted, expected_info_tuples):
        assert one_result.problem_type == ProblemType.VALUE_TYPE_MISMATCH
        assert one_result.res_id == expected_info[0]
        assert one_result.expected == expected_info[1]
        assert one_result.prop_name == expected_info[2]
    assert not _get_validation_status(sorted_problems, is_on_prod=True)
    assert not _get_validation_status(sorted_problems, is_on_prod=False)


def test_unique_value_violation(unique_value_violation: ValidationReportGraphs) -> None:
    assert not unique_value_violation.conforms


def test_file_value_cardinality_violation(file_value_violation: ValidationReportGraphs) -> None:
    assert not file_value_violation.conforms


def test_dsp_inbuilt_violation(dsp_inbuilt_violation: ValidationReportGraphs) -> None:
    assert not dsp_inbuilt_violation.conforms


class TestReformatValidationGraph:
    def test_reformat_unique_value_violation(self, unique_value_violation: ValidationReportGraphs) -> None:
        expected_ids = [
            "identical_values_LinkValue",
            "identical_values_listNode",
            "identical_values_richtext",
            "identical_values_valueAsString",
            "identical_values_valueHas",
        ]
        result = reformat_validation_graph(unique_value_violation)
        sorted_problems = sort_user_problems(result)
        assert len(sorted_problems.unique_violations) == len(expected_ids)
        assert not sorted_problems.user_warnings
        assert not sorted_problems.user_info
        assert not sorted_problems.unexpected_shacl_validation_components
        assert not result.unexpected_results
        alphabetically_sorted = sorted(sorted_problems.unique_violations, key=lambda x: x.res_id)
        for one_result, expected_id in zip(alphabetically_sorted, expected_ids):
            assert one_result.problem_type == ProblemType.DUPLICATE_VALUE
            assert one_result.res_id == expected_id
        assert not _get_validation_status(sorted_problems, is_on_prod=True)
        assert not _get_validation_status(sorted_problems, is_on_prod=False)

    def test_reformat_file_value_violation(self, file_value_violation: ValidationReportGraphs) -> None:
        expected_info_violation = [
            ("authorship_with_newline", ProblemType.GENERIC),
            ("copyright_holder_with_newline", ProblemType.GENERIC),
            ("empty_copyright_holder", ProblemType.INPUT_REGEX),
            ("empty_license", ProblemType.GENERIC),
            ("id_archive_missing", ProblemType.FILE_VALUE),
            ("id_archive_unknown", ProblemType.FILE_VALUE),
            ("id_audio_missing", ProblemType.FILE_VALUE),
            ("id_audio_unknown", ProblemType.FILE_VALUE),
            ("id_document_missing", ProblemType.FILE_VALUE),
            ("id_document_unknown", ProblemType.FILE_VALUE),
            ("id_resource_without_representation", ProblemType.FILE_VALUE_PROHIBITED),
            ("id_still_image_missing", ProblemType.FILE_VALUE),
            ("id_still_image_unknown", ProblemType.FILE_VALUE),
            ("id_text_missing", ProblemType.FILE_VALUE),
            ("id_text_unknown", ProblemType.FILE_VALUE),
            ("id_video_missing", ProblemType.FILE_VALUE),
            ("id_video_unknown", ProblemType.FILE_VALUE),
            ("id_wrong_file_type", ProblemType.FILE_VALUE),
            ("inexistent_license_iri", ProblemType.GENERIC),
            ("license_not_enabled", ProblemType.GENERIC),
            ("unknown_authorship_id", ProblemType.INPUT_REGEX),
        ]
        result = reformat_validation_graph(file_value_violation)
        sorted_problems = sort_user_problems(result)
        alphabetically_sorted_violations = sorted(sorted_problems.unique_violations, key=lambda x: x.res_id)
        assert len(sorted_problems.unique_violations) == len(expected_info_violation)
        assert not sorted_problems.user_warnings
        assert not sorted_problems.user_info
        assert not sorted_problems.unexpected_shacl_validation_components
        assert not result.unexpected_results
        for one_result, expected_info in zip(alphabetically_sorted_violations, expected_info_violation):
            assert one_result.problem_type == expected_info[1]
            assert one_result.res_id == expected_info[0]
        assert not _get_validation_status(sorted_problems, is_on_prod=True)
        assert not _get_validation_status(sorted_problems, is_on_prod=False)

    def test_reformat_dsp_inbuilt_violation(self, dsp_inbuilt_violation: ValidationReportGraphs) -> None:
        expected_info_tuples = [
            ("audio_segment_target_is_video", ProblemType.LINK_TARGET_TYPE_MISMATCH),
            ("audio_segment_target_non_existent", ProblemType.INEXISTENT_LINKED_RESOURCE),
            ("link_obj_target_non_existent", ProblemType.INEXISTENT_LINKED_RESOURCE),
            ("missing_isPartOf", ProblemType.GENERIC),
            ("missing_seqnum", ProblemType.GENERIC),
            ("non_existent_permissions_bitstream", ProblemType.GENERIC),
            ("non_existent_permissions_resource", ProblemType.GENERIC),
            ("non_existent_permissions_value", ProblemType.GENERIC),
            ("region_invalid_geometry", ProblemType.INPUT_REGEX),
            ("region_isRegionOf_resource_does_not_exist", ProblemType.INEXISTENT_LINKED_RESOURCE),
            ("region_isRegionOf_resource_not_a_representation", ProblemType.LINK_TARGET_TYPE_MISMATCH),
            ("target_must_be_a_representation", ProblemType.LINK_TARGET_TYPE_MISMATCH),
            ("target_must_be_an_image_representation", ProblemType.LINK_TARGET_TYPE_MISMATCH),
            ("video_segment_start_larger_than_end", ProblemType.GENERIC),
            ("video_segment_target_is_audio", ProblemType.LINK_TARGET_TYPE_MISMATCH),
            ("video_segment_target_non_existent", ProblemType.INEXISTENT_LINKED_RESOURCE),
            ("video_segment_wrong_bounds", ProblemType.GENERIC),  # once for start that is less than zero
            ("video_segment_wrong_bounds", ProblemType.GENERIC),  # once for the end that is zero
        ]
        result = reformat_validation_graph(dsp_inbuilt_violation)
        sorted_problems = sort_user_problems(result)
        assert len(sorted_problems.unique_violations) == len(expected_info_tuples)
        assert not sorted_problems.user_info
        assert not sorted_problems.unexpected_shacl_validation_components
        assert not result.unexpected_results
        alphabetically_sorted = sorted(sorted_problems.unique_violations, key=lambda x: x.res_id)
        for one_result, expected_info in zip(alphabetically_sorted, expected_info_tuples):
            assert one_result.problem_type == expected_info[1]
            assert one_result.res_id == expected_info[0]
        assert not _get_validation_status(sorted_problems, is_on_prod=True)
        assert not _get_validation_status(sorted_problems, is_on_prod=False)


def test_extract_identifiers_of_resource_results(every_violation_combination_once: ValidationReportGraphs) -> None:
    report_and_onto = every_violation_combination_once.validation_graph + every_violation_combination_once.onto_graph
    data_and_onto = every_violation_combination_once.data_graph + every_violation_combination_once.onto_graph
    result = _extract_base_info_of_resource_results(report_and_onto, data_and_onto)
    result_sorted = sorted(result, key=lambda x: str(x.focus_node_iri))
    expected_iris = [
        (URIRef("http://data/bitstream_no_legal_info"), None),
        (URIRef("http://data/bitstream_no_legal_info"), None),
        (URIRef("http://data/bitstream_no_legal_info"), None),
        (URIRef("http://data/duplicate_still_image_1"), None),
        (URIRef("http://data/duplicate_still_image_2"), None),
        (URIRef("http://data/empty_label"), None),
        (URIRef("http://data/geoname_not_number"), None),
        (URIRef("http://data/id_card_one"), None),
        (URIRef("http://data/id_closed_constraint"), None),
        (URIRef("http://data/id_max_card"), None),
        (URIRef("http://data/id_missing_file_value"), None),
        (URIRef("http://data/identical_values"), None),
        (URIRef("http://data/image_no_legal_info"), None),
        (URIRef("http://data/image_no_legal_info"), None),
        (URIRef("http://data/image_no_legal_info"), None),
        (URIRef("http://data/inexistent_license_iri"), None),
        (URIRef("http://data/label_with_newline"), None),
        (URIRef("http://data/link_target_non_existent"), BNode),
        (URIRef("http://data/link_target_wrong_class"), BNode),
        (URIRef("http://data/list_node_non_existent"), BNode),
        (URIRef("http://data/missing_seqnum"), None),
        (URIRef("http://data/richtext_standoff_link_nonexistent"), None),
        (URIRef("http://data/simpletext_wrong_value_type"), BNode),
        (URIRef("http://data/uri_wrong_value_type"), None),
        (URIRef("http://data/video_segment_start_larger_than_end"), None),
        (URIRef("http://data/video_segment_wrong_bounds"), None),
        (URIRef("http://data/video_segment_wrong_bounds"), None),
    ]
    assert len(result) == len(expected_iris)
    for result_info, expected_iri in zip(result_sorted, expected_iris):
        assert result_info.focus_node_iri == expected_iri[0]
        if expected_iri[1] is None:
            assert not result_info.detail
        else:
            detail_base_info = result_info.detail
            assert isinstance(detail_base_info, DetailBaseInfo)
            assert isinstance(detail_base_info.detail_bn, expected_iri[1])


def test_every_violation_combination_once(every_violation_combination_once: ValidationReportGraphs) -> None:
    assert not every_violation_combination_once.conforms


def test_reformat_every_constraint_once(every_violation_combination_once: ValidationReportGraphs) -> None:
    expected_violations = [
        ("empty_label", ProblemType.INPUT_REGEX),
        ("geoname_not_number", ProblemType.INPUT_REGEX),
        ("id_card_one", ProblemType.MIN_CARD),
        ("id_closed_constraint", ProblemType.NON_EXISTING_CARD),
        ("id_max_card", ProblemType.MAX_CARD),
        ("id_missing_file_value", ProblemType.FILE_VALUE),
        ("identical_values", ProblemType.DUPLICATE_VALUE),
        ("inexistent_license_iri", ProblemType.GENERIC),
        ("label_with_newline", ProblemType.GENERIC),
        ("link_target_non_existent", ProblemType.INEXISTENT_LINKED_RESOURCE),
        ("link_target_wrong_class", ProblemType.LINK_TARGET_TYPE_MISMATCH),
        ("list_node_non_existent", ProblemType.GENERIC),
        ("missing_seqnum", ProblemType.GENERIC),
        ("richtext_standoff_link_nonexistent", ProblemType.GENERIC),
        ("simpletext_wrong_value_type", ProblemType.VALUE_TYPE_MISMATCH),
        ("uri_wrong_value_type", ProblemType.VALUE_TYPE_MISMATCH),
        ("video_segment_start_larger_than_end", ProblemType.GENERIC),
        ("video_segment_wrong_bounds", ProblemType.GENERIC),  # once for start that is less than zero
        ("video_segment_wrong_bounds", ProblemType.GENERIC),  # once for the end that is zero
    ]
    expected_warnings = [
        ("bitstream_no_legal_info", ProblemType.GENERIC),
        ("bitstream_no_legal_info", ProblemType.GENERIC),
        ("bitstream_no_legal_info", ProblemType.GENERIC),
        ("image_no_legal_info", ProblemType.GENERIC),
        ("image_no_legal_info", ProblemType.GENERIC),
        ("image_no_legal_info", ProblemType.GENERIC),
    ]
    expected_info = [
        ("duplicate_still_image_1", ProblemType.FILE_DUPLICATE),
        ("duplicate_still_image_2", ProblemType.FILE_DUPLICATE),
    ]
    result = reformat_validation_graph(every_violation_combination_once)
    sorted_problems = sort_user_problems(result)
    alphabetically_sorted_violations = sorted(sorted_problems.unique_violations, key=lambda x: x.res_id)
    alphabetically_sorted_warnings = sorted(sorted_problems.user_warnings, key=lambda x: x.res_id)
    alphabetically_sorted_info = sorted(sorted_problems.user_info, key=lambda x: x.res_id)
    assert len(sorted_problems.unique_violations) == len(expected_violations)
    assert len(sorted_problems.user_warnings) == len(expected_warnings)
    assert len(sorted_problems.user_info) == len(expected_info)
    assert not sorted_problems.unexpected_shacl_validation_components
    assert not result.unexpected_results
    for one_result, expected in zip(alphabetically_sorted_violations, expected_violations):
        assert one_result.res_id == expected[0]
        assert one_result.problem_type == expected[1]
    for one_result, expected in zip(alphabetically_sorted_warnings, expected_warnings):
        assert one_result.problem_type == expected[1]
        assert one_result.res_id == expected[0]
    for one_result, expected in zip(alphabetically_sorted_info, expected_info):
        assert one_result.problem_type == expected[1]
        assert one_result.res_id == expected[0]
    assert not _get_validation_status(sorted_problems, is_on_prod=True)
    assert not _get_validation_status(sorted_problems, is_on_prod=False)


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

from datetime import datetime
from typing import Any
from typing import cast
from urllib.parse import quote_plus

from loguru import logger
from rdflib import RDF
from rdflib import Graph
from rdflib import URIRef
from tqdm import tqdm

from dsp_tools.clients.connection import Connection
from dsp_tools.commands.xmlupload.iri_resolver import IriResolver
from dsp_tools.commands.xmlupload.make_rdf_graph.jsonld_utils import serialise_jsonld_for_value
from dsp_tools.commands.xmlupload.make_rdf_graph.make_values import make_richtext_value_graph
from dsp_tools.commands.xmlupload.models.upload_state import UploadState
from dsp_tools.commands.xmlupload.stash.stash_models import StandoffStash
from dsp_tools.commands.xmlupload.stash.stash_models import StandoffStashItem
from dsp_tools.commands.xmlupload.stash.stash_models import Stash
from dsp_tools.error.exceptions import BaseError


def upload_stashed_xml_texts(upload_state: UploadState, con: Connection) -> None:
    """
    After all resources are uploaded, the stashed xml texts must be applied to their resources in DSP.
    The upload state is updated accordingly, as a side effect.

    Args:
        upload_state: the current state of the upload
        con: connection to DSP
    """
    logger.info("Upload the stashed XML texts...")
    upload_state.pending_stash = cast(Stash, upload_state.pending_stash)
    standoff_stash = cast(StandoffStash, upload_state.pending_stash.standoff_stash)
    progress_bar = tqdm(
        standoff_stash.res_2_stash_items.copy().items(), desc="Upload stashed XML texts", dynamic_ncols=True
    )
    for res_id, stash_items in progress_bar:
        res_iri = upload_state.iri_resolver.get(res_id)
        if not res_iri:
            # resource could not be uploaded to DSP, so the stash cannot be uploaded either
            # no action necessary: this resource will remain in the list of not uploaded stash items,
            # which will be handled by the caller
            continue
        try:
            resource_in_triplestore = con.get(f"/v2/resources/{quote_plus(res_iri)}")
        except BaseError as err:
            _log_unable_to_retrieve_resource(resource=res_id, received_error=err)
            continue
        logger.info(f"  Upload XML text(s) of resource '{res_id}'...")
        for stash_item in stash_items:
            value_iri = _get_value_iri(stash_item.value.prop_iri, resource_in_triplestore, stash_item.value.value_uuid)
            if not value_iri:
                continue
            if _upload_stash_item(
                stash_item=stash_item,
                res_iri=res_iri,
                value_iri=value_iri,
                iri_resolver=upload_state.iri_resolver,
                con=con,
            ):
                standoff_stash.res_2_stash_items[res_id].remove(stash_item)
        if not standoff_stash.res_2_stash_items[res_id]:
            standoff_stash.res_2_stash_items.pop(res_id)


def _get_value_iri(
    property_name: str,
    resource: dict[str, Any],
    uuid: str,
) -> str | None:
    prefixed_prop = _make_prefixed_prop_from_absolute_iri(property_name)
    values_on_server = resource.get(prefixed_prop)
    if not isinstance(values_on_server, list):
        values_on_server = [values_on_server]

    # get the IRI of the value that contains the UUID in its text
    text_and_iris = ((v["knora-api:textValueAsXml"], v["@id"]) for v in values_on_server)
    value_iri: str | None = next((iri for text, iri in text_and_iris if uuid in text), None)
    # in case that "value_iri" is None, the value that contains the UUID in its text does not exist in DSP
    # no action necessary: this resource will remain in nonapplied_xml_texts,
    # which will be handled by the caller
    return value_iri


def _make_prefixed_prop_from_absolute_iri(absolute_iri: str) -> str:
    _, onto, prop = absolute_iri.rsplit("/", 2)
    local_name = prop.split("#")[-1]
    return f"{onto}:{local_name}"


def _upload_stash_item(
    stash_item: StandoffStashItem,
    res_iri: str,
    value_iri: str,
    iri_resolver: IriResolver,
    con: Connection,
) -> bool:
    """
    Upload a single stashed xml text to DSP.

    Args:
        stash_item: the stashed text value to upload
        res_iri: the iri of the resource
        value_iri: the iri of the value
        iri_resolver: resolver to map ids from the XML file to IRIs in DSP
        con: connection to DSP

    Returns:
        True, if the upload was successful, False otherwise
    """
    payload = _serialise_richtext_for_update(
        stash_item=stash_item,
        value_iri_str=value_iri,
        res_iri_str=res_iri,
        iri_resolver=iri_resolver,
    )
    try:
        con.put(route="/v2/values", data=payload)
    except BaseError as err:
        _log_unable_to_upload_xml_resource(err, stash_item.res_id, stash_item.value.prop_iri)
        return False
    logger.debug(f'  Successfully uploaded xml text of "{stash_item.value.prop_iri}"')
    return True


def _serialise_richtext_for_update(
    stash_item: StandoffStashItem, value_iri_str: str, res_iri_str: str, iri_resolver: IriResolver
) -> dict[str, Any]:
    graph = _make_richtext_update_graph(
        stash_item=stash_item,
        value_iri_str=value_iri_str,
        res_iri_str=res_iri_str,
        iri_resolver=iri_resolver,
    )
    return serialise_jsonld_for_value(graph, res_iri_str)


def _make_richtext_update_graph(
    stash_item: StandoffStashItem, value_iri_str: str, res_iri_str: str, iri_resolver: IriResolver
) -> Graph:
    res_iri = URIRef(res_iri_str)
    value_iri = URIRef(value_iri_str)
    val_graph = make_richtext_value_graph(
        val=stash_item.value,
        val_node=value_iri,
        res_node=res_iri,
        iri_resolver=iri_resolver,
    )
    val_graph.add((res_iri, RDF.type, URIRef(stash_item.res_type)))
    return val_graph


def _log_unable_to_retrieve_resource(
    resource: str,
    received_error: BaseError,
) -> None:
    """
    This function logs the error if it is not possible to retrieve the resource.

    Args:
        resource: the resource id
        received_error: the error
    """
    # print the message to keep track of the cause for the failure
    # apart from that; no action is necessary:
    # this resource will remain in nonapplied_xml_texts, which will be handled by the caller
    err_msg = (
        f"Unable to upload XML texts of resource '{resource}', "
        "because the resource cannot be retrieved from the DSP server."
    )
    print(f"{datetime.now()}:   WARNING: {err_msg} Original error message: {received_error.message}")
    logger.error(err_msg)


def _log_unable_to_upload_xml_resource(
    received_error: BaseError,
    stashed_resource_id: str,
    prop_name: str,
) -> None:
    """
    This function logs if it is not possible to upload a xml resource.

    Args:
        received_error: Error received
        stashed_resource_id: id of the resource
        prop_name: name of the property
    """
    # print the message to keep track of the cause for the failure
    # apart from that; no action is necessary:
    # this resource will remain in nonapplied_xml_texts, which will be handled by the caller
    err_msg = f"Unable to upload the xml text of '{prop_name}' of resource '{stashed_resource_id}'."
    print(f"{datetime.now()}:     WARNING: {err_msg} Original error message: {received_error.message}")
    logger.error(err_msg)



# mypy: disable-error-code="method-assign,no-untyped-def"
import pytest

from dsp_tools.xmllib.helpers import ListLookup


@pytest.fixture
def list_lookup() -> ListLookup:
    return ListLookup.create_new("testdata/json-project/test-list-lookup.json", "en", "default")


class TestGetListLookup:
    def test_lookup(self, list_lookup):
        assert list_lookup._label_language == "en"
        assert set(list_lookup._lookup.keys()) == {"list1", "list2"}
        expected_1 = {
            "label 1": "list1_node1",
            "Label 1": "list1_node1",
            "label 2": "list1_node2",
            "label 2.1": "list1_node2.1",
            "label 2.2": "list1_node2.2",
            "label 2.2.1": "list1_node2.2.1",
        }
        assert list_lookup._lookup["list1"] == expected_1
        expected_2 = {"label 1": "list2_node1", "label 2": "list2_node2"}
        assert list_lookup._lookup["list2"] == expected_2

    def test_prop_to_list_name(self, list_lookup):
        expected = {
            "default:defaultOntoHasListOne": "list1",
            ":defaultOntoHasListOne": "list1",
            "other-onto:otherOntoHasListOne": "list1",
            "other-onto:otherOntoHasListTwo": "list2",
        }
        assert list_lookup._prop_to_list_name == expected
